# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/09/02

Desc: fastspider核心代码, 数据库mysqldb
"""
import functools
from typing import List, Dict

import pymysql
from pymysql import cursors

from fastspider.settings.common import MYSQL_HOST, MYSQL_USERNAME, MYSQL_PASSWORD, MYSQL_DATABASE, MYSQL_PORT
from fastspider.utils.logger import log

from dbutils.pooled_db import PooledDB


def retry(func):
	# 重试机制
	@functools.wraps
	def inner(*args, **kwargs):
		try:
			return func(*args, **kwargs)
		except Exception as e:
			log.error(f"查询数据异常, sql: {kwargs.get('sql')}, error: {e}")
		return func(*args, **kwargs)

	return inner


class MysqlDB(object):

	def __init__(self, host=None, port=3306, username=None, password=None, database=None, **kwargs):
		self.host = host if host else MYSQL_HOST
		self.port = port if port else MYSQL_PORT
		self.username = username if username else MYSQL_USERNAME
		self.password = password if password else MYSQL_PASSWORD
		self.database = database if database else MYSQL_DATABASE

		# DBUtils 是一套允许线程化 Python 程序可以安全和有效的访问数据库的模块。使用该模块实现数据库mysql的链接
		try:
			self.connect_pooled_db = PooledDB(
				creator=pymysql,
				host=self.host,
				port=self.port,
				user=self.username,
				password=self.password,
				db=self.database,
				charset="utf8mb4",
				cursorclass=cursors.SSCursor,
				mincached=1,
				maxcached=100,
				maxconnections=100,
				blocking=True,
				ping=7
			)
		except Exception as e:
			log.error(
				f"数据库链接【失败】, 账号信息: username: {self.username}, password: {self.password}, host: {self.host}, port: {self.port}, database: {self.database}, 详细失败原因: {e}")
		else:
			log.debug(
				f"数据库链接【成功】, 账号信息: username: {self.username}, host: {self.host}, port: {self.port}, database: {self.database}")

	def get_connect(self):
		"""
			获取数据库链接
		:return: 数据库链接对象, 游标
		"""
		conn = self.connect_pooled_db.connection(shareable=False)
		cursor = conn.cursor()
		return conn, cursor

	def close(self, conn, cursor):
		"""
			关闭数据库链接
		:param conn: 数据库链接对象
		:param cursor: 游标
		"""
		conn.close()
		cursor.close()

	# 封装方法
	@retry
	def find(self, sql, limit=0):
		"""
			查询数据
		:param sql:  查询sql语句
		:param limit:  限制查询的结果, 0表示查询所有数据 1表示查询1条数据 其他值表示查询指定条数数据
		:return:
		"""
		conn, cursor = self.get_connect()
		cursor.execute(sql)
		if limit == 0:
			result = cursor.fetchall()
		elif limit == 1:
			result = cursor.fetchone()
		else:
			result = cursor.fetchmany(limit)

		self.close(conn, cursor)
		return result

	def insert(self, sql):
		"""
			插入数据
		:param sql: 插入sql语句
		:return: 插入数据成功后, 受影响的条数
		"""
		affect_rows = None
		try:
			conn, cursor = self.get_connect()
			affect_rows = cursor.execute(sql)
			conn.commit()
		except Exception as e:
			log.error(f"插入数据失败, sql: {sql}， error: {e}")
		finally:
			self.close(conn, cursor)

		return affect_rows

	def insert_many(self, sql, data: List[Dict]):
		"""
			批量插入数据, 性能比insert强
		:param sql: sql语句模板。如 insert into table (files1, fields2, files3....) values (value1, value2, value3....)
		:param data: 插入的数据。如[{1,2,3}, {4,5,6} .... ]
		:return: 插入数据成功后, 受影响的条数
		"""
		affect_rows = None
		try:
			conn, cursor = self.get_connect()
			affect_rows = cursor.executemany(sql, data)
			conn.commit()
		except Exception as e:
			log.error(f"批量插入数据失败, sql: {sql}， error: {e}")
		finally:
			self.close(conn, cursor)

		return affect_rows

	def update(self, sql):
		"""
			更新数据
		:param sql: 更新sql语句
		:return: True表示更新成功  False表示更新失败
		"""
		try:
			conn, cursor = self.get_connect()
			cursor.execute(sql)
			conn.commit()
		except Exception as e:
			log.error(f"更新数据失败, sql: {sql}， error: {e}")
			return False
		else:
			return True
		finally:
			self.close(conn, cursor)

	def delete(self, sql):
		"""
			删除数据
		:param sql: 删除sql语句
		:return: True表示更新成功  False表示更新失败
		"""
		try:
			conn, cursor = self.get_connect()
			cursor.execute(sql)
			conn.commit()
		except Exception as e:
			log.error(f"更新数据失败, sql: {sql}, error: {e}")
			return False
		else:
			return True
		finally:
			self.close(conn, cursor)

	def execute(self, sql):
		"""
			执行sql
		:param sql: sql语句
		:return: True表示更新成功  False表示更新失败
		"""
		try:
			conn, cursor = self.get_connect()
			cursor.execute(sql)
			conn.commit()
		except Exception as e:
			log.error(f"执行sql语句失败, sql: {sql}， error: {e}")
			return False
		else:
			return True
		finally:
			self.close(conn, cursor)


# encoding=utf-8
import datetime
import pickle
import time

from fastspider.settings import common
from fastspider.utils.logger import log


def key_to_hump(key):
	"""
		将变量转化为驼峰命名
	:return: 返回驼峰命名变量
	"""
	return key.lower().title().replace("_", "")


def check_class_method(obj, name):
	"""
		检查对象是否有某个属性
	:return:
	"""
	try:
		return getattr(obj, str(name))
	except Exception as e:
		print(f"对象{obj} 没有方法{str(name)}")
		return None


def get_tunnel_proxy():
	"""
		隧道代理
	"""
	try:
		if common.PROXY_ENABLE:
			proxy_meta = "http://%(user)s:%(pass)s@%(host)s:%(port)s" % {
				"host": common.PROXY_TUNNEL_HOST,
				"port": common.PROXY_TUNNEL_PORT,
				"user": common.PROXY_TUNNEL_USER,
				"pass": common.PROXY_TUNNEL_PASSWORD,
			}
			proxies = {
				"http": proxy_meta,
				"https": proxy_meta,
			}
			return proxies
		return None
	except Exception as e:
		return None


def sleep_time(times):
	"""
		休眠指定时间
	:param times: 秒
	"""
	time.sleep(times)


class Singleton(object):
	def __init__(self, cls):
		self._cls = cls
		self._instance = {}

	def __call__(self, *args, **kwargs):
		if self._cls not in self._instance:
			self._instance[self._cls] = self._cls(*args, **kwargs)
		return self._instance[self._cls]


def cookies2dict(cookies):
	"""
		将request中的cookies解析为字典
	"""
	cookie_dict = {}
	for i in cookies.split(";"):
		cookie = i.split("=")
		cookie_dict[cookie[0]] = cookie[1]
	return cookie_dict


def pascal_case_to_snake_case(name):
	"""
		将驼峰命名改为单词下划线命名
	:param name: 驼峰命名变量
	:return: 单词下划线命名变量
	"""
	snake_name = []
	for index, char in enumerate(name):
		if index != 0 and char.isupper():
			snake_name.append("_")
		snake_name.append(char)
	return "".join(snake_name).lower()


def time_to_timestamp(strtime):
	"""
		将字符串时间转换为时间戳
	"""
	timeArray = time.strptime(strtime, "%Y-%m-%d %H:%M")
	return int(time.mktime(timeArray))


def get_current_timestamp():
	"""
		获取当前时间的时间戳
	"""
	return int(time.time())


def format_date(format="%Y-%m-%d %H:%M:%S"):
	"""
		格式化日期
	"""
	return datetime.datetime.now().strftime(format)


def dumps_obj(obj):
	return pickle.dumps(obj)


def loads_obj(obj_str):
	return pickle.loads(obj_str)


def get_method(obj, name):
	"""
		获取对象的可执行方法
	"""
	try:
		return getattr(obj, str(name))
	except AttributeError as e:
		log.error("Method %r not found in: %s" % (name, obj))
		return None



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/07/30

Desc: fastspider核心代码, 实体Item
"""


class BaseItemMetaClass(type):

	def __new__(cls, name, bases, attrs):
		attrs.setdefault("__name__", None)
		attrs.setdefault("__table_name__", None)
		attrs.setdefault("__update_key__", None)
		attrs.setdefault("__unique_key__", None)
		return type.__new__(cls, name, bases, attrs)


class Item(metaclass=BaseItemMetaClass):
	"""
		定义继承的元类, 子类直接拥有元类中的属性。
	"""

	def __init__(self):
		pass

	def __setitem__(self, key, value):
		self.__class__.__dict__[key] = value

	def __getitem__(self, key):
		return self.__class__.__dict__[key]

	@property
	def item_name(self):
		return self.__class__.__name__

	@property
	def table_name(self):
		return self.__table_name__

	@table_name.setter
	def table_name(self, name):
		self.__table_name__ = name

	@property
	def to_dict(self):
		item_property = {}
		for key, values in self.__dict__.items():
			if key not in ("__name__", "__table_name__", "__update_key__", "__unique_key__"):
				item_property[key] = values
		return item_property



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/07/29

Desc: 配置文件
"""

# 爬虫配置
SPIDER_THREAD_COUNT = 1  # 爬虫并发数

# 爬虫相关
# 分布式爬虫使用到的参数
COLLECTOR_SLEEP_TIME = 1  # 从redis任务队列中获取任务到内存队列的间隔
COLLECTOR_TASK_COUNT = 10  # 每次获取任务数量

# request防丢机制。（指定的REQUEST_LOST_TIMEOUT时间内request还没做完，会重新下发 重做）
REQUEST_LOST_TIMEOUT = 600  # 10分钟
# 每个parser从内存队列中获取任务的数量
SPIDER_TASK_COUNT = 1

# request网络请求超时时间
REQUEST_TIMEOUT = 22  # 等待服务器响应的超时时间，浮点数，或(connect timeout, read timeout)元组

# UserAgent类型 支持 'chrome', 'opera', 'firefox', 'internetexplorer', 'safari'，若不指定则随机类型
USER_AGENT_TYPE = "chrome"

# requests 使用session
USE_SESSION = False

# 下载时间间隔 单位秒。
# 支持格式, 如 SPIDER_SLEEP_TIME = 3
# 如 SPIDER_SLEEP_TIME = [2, 5] 或者 (2, 5) 则间隔为 2~5秒之间的随机数，包含2和5
SPIDER_SLEEP_TIME = (
	1, 2
)

# 是否开启代理
PROXY_ENABLE = True
# 是否使用隧道代理
PROXY_TUNNEL_HOST = None
PROXY_TUNNEL_PORT = None
PROXY_TUNNEL_USER = None
PROXY_TUNNEL_PASSWORD = None

# 浏览器渲染
WEBDRIVER = dict(
	not_load_images=True,  # 是否加载图片
	user_agent=None,  # 字符串 或 无参函数，返回值为user_agent
	proxies=None,  # xxx.xxx.xxx.xxx:xxxx 或 无参函数，返回值为代理地址
	headless=False,  # 是否为无头浏览器
	driver_type="Chrome",  # CHROME、PHANTOMJS、FIREFOX
	timeout=30,  # 请求超时时间
	window_size=(1024, 800),  # 窗口大小
)

# 数据入库的pipeline，可自定义，默认RedisPipeline
ITEM_PIPELINES = [
	# "fastspider.pipeline.test.TmallCheapPipeline"
	# "fastspider.pipeline.redis_pipeline.RedisPipeline",
	# "fastspider.pipeline.mysql_pipeline.MysqlPipeline",
	# "fastspider.pipeline.mongo_pipeline.MongoPipeline",
]

# 日志配置
LOGGER = dict(
	# LOG_IS_WRITE_TO_CONSOLE=True,  # 是否打印到控制台
	# LOG_NAME=os.path.basename(os.getcwd()),
	# LOG_PATH="log/%s.log",  # log存储路径
	# LOG_LEVEL="DEBUG",
	# LOG_IS_WRITE_TO_FILE=True,  # 是否保存到文件中
	# LOG_MAX_BYTES=10 * 1024 * 1024,  # 每个日志文件的最大字节数
	# LOG_FILE_MAX_NUMBER=10,  # 最大备份文件个数
	# LOG_ENCODING="utf8",  # 日志文件编码
)

# 去重
ITEM_FILTER_ENABLE = False  # item 去重

# Redis中表的配置
REDIS_KEY = "fastspider"
# 爬虫总任务表
REDIS_MISSION_REQUESTS = "{redis_key}:f_all_requests"
# 爬虫失败请求任务表
REDIS_MISSION_FAIL_REQUESTS = "{redis_key}:f_fail_requests"
# 爬虫状态任务表
REDIS_SPIDER_STATUS = "{redis_key}:f_spider_status"

# 【必填项】 REDIS链接
REDISDB_URL = ""
REDISDB_IP = "127.0.0.1"
REDISDB_PORT = 6379
REDISDB_USER_PASS = ""
# 默认 0 到 15 共16个数据库
REDISDB_DB = 0
# 心跳检测的数据缓存的时间，以秒为单位
REDISDB_TIME = 5

# 【必填项】MYSQL
MYSQL_HOST = "127.0.0.1"
MYSQL_PORT = 3306
MYSQL_DATABASE = "dc_manage"
MYSQL_USERNAME = "root"
MYSQL_PASSWORD = "123123"

# 预警/消息通知

# 飞书自定义机器人 webhook地址
FEISHU_WEB_HOOK = ''
# 飞书自定义机器人 签名校验参数
FEISHU_SECRET = ''


# 导入自定义settings的配置文件
try:
	from settings import *
except Exception as e:
	pass



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/08/25

Desc: 解析启动爬虫命令行参数, 并启动对应的爬虫
"""
import importlib


class RunFastSpider(object):

	@classmethod
	def run(cls, spider_path, thread_count):
		module, class_name = spider_path.rsplit(".", 1)
		mod = importlib.import_module(module)
		try:
			obj = getattr(mod, class_name)
		except AttributeError:
			raise NameError(f"Module '{module}' doesn't define any object named '{class_name}'")

		obj(thread_count).start()



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/07/29

Desc: fastspider核心代码, redis缓存
"""
from threading import Thread

from fastspider.settings import common
from fastspider.utils.logger import log

from fastspider.db.redis_db import RedisDB
from fastspider.utils import tools


class RedisCache(Thread):

	def __init__(self, parser_name):
		super(RedisCache, self).__init__()

		self._thread_stop = False
		self._redis = RedisDB()
		self._parser_name = parser_name
		self._redis_time = common.REDISDB_TIME

	def heartbeat(self, parser_name):
		"""
			心跳检测
		"""
		log.debug("---%s 心跳检测---" % parser_name)
		self._redis.set(parser_name, 1)
		self._redis.set_expire(parser_name,self._redis_time)

	def run(self):
		while not self._thread_stop:
			for parser_name in self._parser_name:
				self.heartbeat(parser_name)
			tools.sleep_time(1)

	def stop(self):
		self._thread_stop = True



from queue import PriorityQueue


class Test():

	def __init__(self):
		self.priority = 300

	def __lt__(self, other):
		return self.priority < other.priority


queue = PriorityQueue()

for i in range(5):
	t = Test()
	queue.put(t)

print(queue.qsize())
print(queue.get())



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/07/30

Desc: fastspider核心代码, 封装requests包
"""
import requests
import urllib3
import copy
from requests.adapters import HTTPAdapter

from fastspider.http import user_agent
from fastspider.http.fastselenium.fastwebdriver import WebDriverPool
from fastspider.http.response.response import Response
from fastspider.settings import common
from fastspider.utils import tools

# 忽略警告信息
from fastspider.utils.logger import log

urllib3.disable_warnings()

urllib3.util.ssl_.DEFAULT_CIPHERS += ':RC4-SHA'


class Request(object):
	user_agent_pool = user_agent
	proxies_pool = None
	webdriver_pool = None
	session = None

	# 组装非必须传入的参数
	__request_attrs__ = ["headers", "data", "params", "cookies", "json", "timeout", "proxies", "verify"]
	# 组装默认传递的参数
	__default_requests_attrs__ = {
		"url": "",
		"method": "",
		"retry_time": 0,
		"priority": 300,
		"parser_name": None,
		"callback": None,
		"filter_request": False,
		"use_session": False,
		"download_middleware": None,
		"web_render": False,
		"web_render_time": 0,
		"web_render_scroll": False,
		"request_sync": False,
		"meta": None,
		"sleep_time": 0
	}

	def __init__(self, url="", method="", retry_time=0, priority=300, parser_name=None, callback=None,
	             filter_request=False, use_session=False, download_middleware=None, web_render=False, web_render_time=0,
	             web_render_scroll=False, request_sync=False, meta=None, sleep_time=0, **kwargs):
		self.url = url
		self.method = method
		self.retry_time = retry_time
		self.priority = priority
		self.parser_name = parser_name
		self.callback = callback
		self.filter_request = filter_request
		self.request_sync = request_sync
		self.use_session = use_session
		self.download_middleware = download_middleware
		self.web_render = web_render
		self.web_render_time = web_render_time
		self.web_render_scroll = web_render_scroll
		# meta参数的深拷贝
		self.meta = dict(meta) if meta else {}
		self.sleep_time = sleep_time

		self.request_kwargs = {}
		for key, value in kwargs.items():
			if key in self.__class__.__request_attrs__:
				self.request_kwargs[key] = value

			self.__dict__[key] = value

	def __lt__(self, other):
		"""
			把Request对象插入到优先级对象中, 会判断每一个Reqeest对象的优先级。
			需要重定义__lt__方法, 使类能进行大小比较, 然后将具有优先级的Request对象插入到优先级队列中
		:param other:
		:return:
		"""
		return self.priority < other.priority

	def __repr__(self):

		return f"<Request {self.url}>"

	@property
	def to_dict(self):
		request_dict = {}

		for key, value in self.__dict__.items():
			if key == "callback":
				request_dict[key] = getattr(value, "__name__") if value else "parser"
				continue

			request_dict[key] = value

		return request_dict

	@classmethod
	def from_dict(cls, request_dict):
		for key, value in request_dict.items():
			if isinstance(value, bytes):
				request_dict[key] = tools.loads_obj(value)
		return cls(**request_dict)

	@property
	def _make_session(self):
		"""
		:return: 生成request.Session对象
		"""
		self.__class__.session = requests.Session()
		# pool_connections – 缓存的 urllib3 连接池个数  pool_maxsize – 连接池中保存的最大连接数
		adapter = HTTPAdapter(pool_connections=500, pool_maxsize=500)
		# mount 挂载前缀, 只要URL是以http前缀开头的会话, 都会被传入到适配器adapter中
		self.__class__.session.mount("http", adapter)
		return self.__class__.session

	@property
	def _webdriver_pool(self):
		if not self.__class__.webdriver_pool:
			self.__class__.webdriver_pool = WebDriverPool(**common.WEBDRIVER)
		return self.__class__.webdriver_pool

	def get_response(self):
		# 获取meta参数
		meta = self.meta

		# 设置请求超时时间
		self.request_kwargs.setdefault(
			"timeout", common.REQUEST_TIMEOUT
		)

		# 设置ssl验证
		self.request_kwargs.setdefault(
			"verify", False
		)

		# 获取请求方式method
		method = "GET"
		if self.__dict__.get("method"):
			method = self.__dict__.get("method")
		elif "data" in self.request_kwargs:
			method = "POST"
		# 设置ua
		headers = self.request_kwargs.get("headers", {})
		if "user-agent" not in headers and "User-Agent" not in headers:
			headers["User-Agent"] = self.__class__.user_agent_pool.get_ua(common.USER_AGENT_TYPE)
			self.request_kwargs.update(headers=headers)
		else:
			if "user-agent" in headers:
				headers.pop("user-agent")
			headers.setdefault(
				"User-Agent", self.__class__.user_agent_pool.get_ua(common.USER_AGENT_TYPE)
			)
			self.request_kwargs.update(headers=headers)

		# 代理proxies
		proxies = self.request_kwargs.get("proxies", False)
		# if proxies and common.PROXY_ENABLE:
		# 	# 代理的获取方式, 依次获取隧道代理、IP地址代理.....
		# 	abuyun_proxies = tools.get_tunnel_proxy()
		# 	if abuyun_proxies:
		# 		self.request_kwargs.update(proxies=abuyun_proxies)
		self.request_kwargs.update(proxies=proxies)

		# TODO: 如果没有隧道代理, 则可以使用IP代理

		# 浏览器渲染
		use_session = self.use_session if self.use_session else common.USE_SESSION

		if self.web_render:
			try:
				driver = self._webdriver_pool.get()

				driver.get(self.url)

				if self.web_render_scroll:
					js = "return action=document.body.scrollHeight"
					height = driver.execute_script(js)
					for i in range(0, height, 150):
						if self.web_render_time:
							tools.sleep_time(self.web_render_time)
						driver.execute_script(f"window.scrollTo(0, {i})")

				if self.web_render_time:
					tools.sleep_time(self.web_render_time)

				html = driver.page_source
				response = Response.from_dict({
					"status_code": 200,
					"_content": html,
					"url": self.url,
					"cookies": driver.cookies,
					"meta": meta
				})
				response.driver = driver
			except Exception as e:
				print(e)
				raise e
			finally:
				self._webdriver_pool.remove(driver)

		# 设置session
		elif use_session:
			response = self._make_session.request(method=method, url=self.url, **self.request_kwargs)
			response.__dict__.setdefault("meta", meta)
			response = Response(response)
		else:
			response = requests.request(method=method, url=self.url, **self.request_kwargs)
			response.__dict__.setdefault("meta", meta)
			response = Response(response)

		log.debug(
			"""
			------------------%s.%s request for-----------------------------
				url = %s
				method = %s
				request_kwargs = %s
			""" % (
				self.parser_name,
				self.callback,
				self.url,
				method,
				self.request_kwargs,
			)
		)

		return response



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/07/30

Desc: fastspider核心代码, 封装response包
"""
from bs4 import BeautifulSoup
from requests.cookies import RequestsCookieJar
from requests.models import Response as res

# from fastspider.http.selector import Selector
from parsel import Selector


class Response(res):
	"""
		封装requests中的响应Response
	"""

	def __init__(self, response):
		super(Response, self).__init__()
		# 将response的属性同步到封装的Response对象上
		self.__dict__.update(response.__dict__)

		# 编码的严格程度, strict \ replacce \ ignore
		self.encoding_errors = "strict"
		# 加载响应内容
		self._load_selector = None

	@property
	def selector(self):
		"""
			获取已经解析过的selector对象
		:return:
		"""
		if not self._load_selector:
			# TODO: 可以封装Selector的方法
			# self._load_selector = Selector(self.text)
			self._load_selector = Selector(self.text)
		return self._load_selector

	@classmethod
	def from_dict(cls, response_dict):
		"""
			组装响应response
		:return:
		"""
		# 设置cookie
		cookie_jar = RequestsCookieJar()
		cookie_jar.update(other=response_dict["cookies"])
		response_dict["cookies"] = cookie_jar

		response = res()
		response.__dict__.update(response_dict)
		return response

	def bs4(self, parser="html.parser"):
		"""
			使用bs4进行响应内容的解析与提取
		:param parser: 解析器, 可以定义为html.parser、lxml、html5lib
		:return:
		"""
		return BeautifulSoup(self.text, parser)

	def xpath(self, query, **kwargs):
		return self.selector.xpath(query, **kwargs)



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/07/30

Desc: fastspider核心代码, 选择器的封装
"""

from parsel import Selector as ParselSelector


class Selector(ParselSelector):

	def __init__(self, text=None):
		self.text = text
		
		super(Selector, self).__init__(text)




# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/07/29

Desc: fastspider核心代码, 内存db
"""

from queue import PriorityQueue


class MemoryDB(object):
	"""
		PriorityQueue 优先级队列
		格式：q.put((数字,值))
		特点：数字越小，优先级越高。如果不传入数字, 那就按照值的asc值来判断优先级。asc值可以用过ord()方法来获取
	"""

	def __init__(self):
		self.priority_queue = PriorityQueue()

	def put(self, task):
		"""
			添加任务
		"""
		self.priority_queue.put(task)

	def get_nowait(self):
		"""
			获取队列中的任务
			get_nowait(): 强取任务, 获取不到任务, 则出现queue.empty异常
		:return: 返回任务或者None
		"""
		try:
			task = self.priority_queue.get_nowait()
			return task
		except Exception as e:
			return None

	def is_empty(self):
		"""
			判断当前队列是否为空
		:return: 队列为空, 返回True。队列不为空, 返回False
		"""
		return self.priority_queue.empty()



# encoding=utf-8
import time

import fastspider


class TestSpider(fastspider.Spider):
	start_urls = ["https://movie.douban.com/top250", "https://movie.douban.com/top250?start=25&filter=",
	              "https://movie.douban.com/top250?start=50&filter=", "https://movie.douban.com/top250?start=75&filter=",
	              "https://movie.douban.com/top250?start=100&filter="]

	def start_requests(self):
		for url in self.start_urls:
			yield fastspider.Request(url=url)

	def parser(self, request, response):
		movies = response.xpath('//*[@id="content"]/div/div[1]/ol/li')
		for movie in movies:
			href = movie.xpath('./div/div[2]/div[1]/a/@href')[0].get()
			title = movie.xpath('./div/div[2]/div[1]/a/span[1]/text()')[0].get()

			print(href, title)


if __name__ == "__main__":
	TestSpider("test", 3).start()



import time
import hashlib
import requests

headers = {
	'Host': 'h5api.m.taobao.com',
	'cookie': 't=af7b753193ca2feb6387378709a363cc; _tb_token_=eb1ee3e1ee935; cookie2=1bbde7bef2d2eb9e2ea6c9104b2bbf8d; cna=PeiLFJrhqW4CAXkgmv3+E70E; _m_h5_tk=0f102b40794f5ce1cbf6189efee84f38_1543985837478; _m_h5_tk_enc=c6b00d0f38cdca735cc190e30f9eadf1; isg=BDQ0YzZQ1GD2NECvhLnjwVQ7BfJmpR7rLc7aM86VwL9COdSD9h0oh-r7vTdE2pBP',
	'Connection': 'keep-alive',
	'f-refer': 'wv_h5',
	'Accept': '*/*',
}

data = '{"shopId":"586728252","sellerId":"3459177618"}'  # 必须为这个格式
url = 'https://h5api.m.taobao.com/h5/mtop.taobao.shop.wireless.category.get/1.0/?jsv=2.4.5&appKey=12574478&t={time}&sign={sign}&api=mtop.taobao.shop.wireless.category.get&v=1.0&H5Request=true&preventFallback=true&type=jsonp&dataType=jsonp&callback=mtopjsonp2&data={data}'

# 第一次请求
# 获取
# _m_h5_tk
response = requests.get(url.format(sign='a2b39c32169159812ee93f0f4752e017', data=data, time=int(time.time() * 1000)))
res = response.cookies.get_dict()
print(res)
s = res['_m_h5_tk']
# _m_h5_tk
# 用于sign
# 的参数
s = s.split('_')[0]
print('_m_h5_tk:', res)
r = ';'.join(['{}={}'.format(k, v) for k, v in res.items()])  # 拼接ｃｏｏｏｋｉｅ
print(r)
headers['cookie'] = r

a = '{}&{}&12574478&{}'.format(s, int(time.time() * 1000), data)
print('params:', a)
print('t', int(time.time() * 1000))
ss = hashlib.md5(a.encode())
print('sign', ss.hexdigest())

# 发第二次请求
response = requests.get(url.format(sign=ss.hexdigest(), data=data, time=int(time.time() * 1000)), headers=headers)
print(response.content.decode())
print(response.url)
print(response.text)



import redis

r = redis.StrictRedis(host="127.0.0.1", port=6379, db=1, password="", decode_responses=True)

pipeline = r.pipeline()
pipeline.multi()

pipeline.execute_command("ZADD", "test", 30, "wang")
a = pipeline.execute()
print(a)
print(type(a))



from fastspider.item.item import Item

item = Item()
item.title = "123"
item.src = "http://dddd"

print(item.title)
print(item.src)


class MyItem(Item):

	def __init__(self, title, src):
		self.title = title
		self.src = src


i = MyItem("121212", "sdasdsd")
print(i.title)
print(i.src)
print(i.table_name)



# encoding=utf-8
import time

import fastspider


class TestSpider(fastspider.LightSpider):
	douban_url = "https://movie.douban.com/top250"

	def start_requests(self):
		yield fastspider.Request(url=self.douban_url)
		# yield fastspider.Request(url=self.douban_url)

	def parser(self, request, response):
		movies = response.xpath('//*[@id="content"]/div/div[1]/ol/li')
		for movie in movies:
			href = movie.xpath('./div/div[2]/div[1]/a/@href')[0].get()
			title = movie.xpath('./div/div[2]/div[1]/a/span[1]/text()')[0].get()

			print(href, title)


if __name__ == "__main__":
	TestSpider(3).start()



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/09/02

Desc: fastspider核心代码, 分布式爬虫spider
"""
from fastspider.core.base.cycle_base import CycleBase
from fastspider.core.scheduler.scheduler import Scheduler
from fastspider.utils import tools
from fastspider.utils.logger import log


class Spider(CycleBase, Scheduler):

	def __init__(self, redis_key=None, thread_count=None, check_task_interval=5):
		"""
			初始化
		"""
		super(Spider, self).__init__(
			redis_key=redis_key,
			thread_count=thread_count,
			check_task_interval=check_task_interval
		)

	def run(self):
		"""
			启动爬虫
		"""
		try:
			if not self._parsers:
				self._parsers.append(self)

			# 开始启动
			self._start()
			# 判断是否结束
			while True:
				if self.all_thread_is_done():
					self.all_thread_stop()

					log.info("无任务, 爬虫执行完毕")
					break

				# 休息1秒后再次检查爬虫是否还在运行
				tools.sleep_time(1)
		except Exception as e:
			log.error(f"爬虫执行异常, 异常原因: {e}")



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/09/02

Desc: fastspider核心代码, 请求request缓存
"""

import collections
import threading

from fastspider.db.redis_db import RedisDB
from fastspider.utils import tools
from fastspider.settings import common
from fastspider.utils.logger import log


class RequestCache(threading.Thread):
	MAX_REQUESTS_COUNT = 1000

	def __init__(self, redis_key):
		super(RequestCache, self).__init__()

		self._thread_stop = False
		self._add_request_to_db_status = False

		self._requests_deque = collections.deque()
		self._del_requests_deque = collections.deque()

		self._redis_db = RedisDB()
		self._mission_request_name = common.REDIS_MISSION_REQUESTS.format(redis_key=redis_key)
		self._mission_fail_request_name = common.REDIS_MISSION_FAIL_REQUESTS.format(redis_key=redis_key)

	def run(self):
		while not self._thread_stop:
			try:
				self.__add_request_to_db()
			except Exception as e:
				log.error(f"向redis中添加待执行任务失败, 失败原因: {e}")
			tools.sleep_time(1)

	def flush(self):
		"""
			主动刷新任务
		"""
		try:
			self.__add_request_to_db()
		except Exception as e:
			log.error(f"向redis中添加待执行任务失败, 失败原因: {e}")

	def stop(self):
		self._thread_stop = True

	def add_request(self, request):
		"""
			向请求request队列中加入请求
		:param request: 请求
		"""
		self._requests_deque.append(request)

		if self.get_request_count() > self.MAX_REQUESTS_COUNT:
			self.flush()

	def get_request_count(self):
		"""
			获取待执行队列中的任务数
		"""
		return len(self._requests_deque)

	def add_del_request(self, request):
		"""
			向队列中添加需要移除的请求request
		:param request: 请求
		"""
		self._del_requests_deque.append(request)

	def add_fail_request(self, request):
		"""
			向队列中添加失败的请求
		:param request:
		:return:
		"""
		try:
			self._redis_db.zadd(self._mission_fail_request_name, request.to_dict, request.priority)
		except Exception as e:
			log.error(f"添加失败请求到redis表{self._mission_fail_request_name}中失败, 错误原因: {e}")

	def __add_request_to_db(self):
		request_list = []
		priority_list = []
		callbacks = []
		while self._requests_deque:
			request = self._requests_deque.popleft()
			self._add_request_to_db_status = True

			if callable(request):
				callbacks.append(request)

			# 请求优先等级
			priority = request.priority
			# TODO: 请求去重, 暂时不考虑
			filter_request = request.filter_request

			request_list.append(str(request.to_dict))
			priority_list.append(priority)

			if len(request_list) > self.__class__.MAX_REQUESTS_COUNT:
				self._redis_db.zadd(self._mission_request_name, request_list, priority_list)
				request_list = []
				priority_list = []

		# 数量达不到 MAX_REQUESTS_COUNT, 则单独处理
		if request_list:
			self._redis_db.zadd(self._mission_request_name, request_list, priority_list)

		# TODO: 不太明白
		# for callback in callbacks:
		# 	callback()

		# 将完成的任务从redis中移除
		if self._del_requests_deque:
			del_requests_list = []
			while self._del_requests_deque:
				del_requests_list.append(self._del_requests_deque.popleft())

			if del_requests_list:
				self._redis_db.zrem(self._mission_request_name, del_requests_list)

		self._add_request_to_db_status = False

	def is_adding_request_to_db(self):
		"""
			获取线程是否还在向db中添加数据
		"""
		return self._add_request_to_db_status



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/07/29

Desc: fastspider核心代码, 控制器
"""
import time
from collections import Iterable
import random
from threading import Thread

from fastspider.http.request.request import Request
from fastspider.item.item import Item
from fastspider.settings import common
from fastspider.utils.logger import log
from fastspider.utils import tools


class BaseController(Thread):
	# 统计任务执行成功的个数 和 任务执行失败的个数, 用于后期统计任务执行成功率, 成功率太低, 则发送报警信息
	# 成功率计算公司  失败任务总数 / 总任务数 > 0.5 则发送报警信息
	_success_task_num = 0
	_fail_task_num = 0

	has_task_flag = False

	def __init__(self, collector=None):
		super(BaseController, self).__init__()

		self._thread_stop = False
		self._parser = []

		self._collector = collector

	def has_task(self):
		return self.has_task_flag

	def stop(self):
		# 将线程启动的标识 设置为True
		self._thread_stop = True

	def add_parser(self, parser):
		self._parser.append(parser)


class LightSpiderController(BaseController):

	def __init__(self, memory_db, item_cache):
		super(LightSpiderController, self).__init__()

		self._memory_db = memory_db
		self._item_cache = item_cache

	def run(self):

		while not self._thread_stop:
			try:
				request = self._memory_db.get_nowait()
				if not request:
					self.has_task_flag = False
				else:
					self.has_task_flag = True
					self.deal_requests([request])
			except Exception as e:
				log.error(f"轻量级爬虫lightspider执行失败, 失败原因: {e}")

	def deal_requests(self, requests):
		"""
			处理请求
		:param request: 请求
		:return:
		"""
		response = None
		for request in requests:
			for parser in self._parser:
				# 判断当前请求是否是请求对应的爬虫触发的。避免同时多个爬虫一起执行时, 请求request和对应的爬虫匹配错误
				if request.parser_name == parser.name:
					if request.sleep_time:
						tools.sleep_time(request.sleep_time)

					try:
						# 解析request
						# if request.download_middleware:
						# 	pass
						# else:
						response = request.get_response()
						# 判断是否有回调函数
						if request.callback:
							# 检查回调函数是否可用
							# callback_parser = (
							# 	request.callback if callable(request.callback)
							# 	else tools.check_class_method(parser, request.callback)
							# )
							callback_parser = (request.callback)
							results = callback_parser(request, response)
						else:
							results = parser.parser(request, response)

						if results and not isinstance(results, Iterable):
							raise Exception(f"{parser.name}.{request.callback}必须可迭代的返回值")

						for result in results or []:
							if isinstance(result, Request):
								result.parser_name = result.parser_name or parser.name
								# 如果是同步的callback, 将解析request对象添加到requests中
								if result.request_sync:
									requests.append(result)
								else:
									# 异步, 将任务添加到 任务队列中
									self._memory_db.put(result)
							elif isinstance(result, Item):
								# 存入item缓存中
								self._item_cache.put(result)

					except Exception as e:
						# TODO: 记录任务失败的信息
						print(e)
						self._memory_db.put(request)
					# if request.retry_time:
					# 	request.retry_time -= 1
					# 	self._memory_db.put(request)
					# finally:
					# 	print("释放相关的链接, 如数据库、浏览器的链接")

					break
		# 休眠
		if common.SPIDER_SLEEP_TIME:
			if isinstance(common.SPIDER_SLEEP_TIME, (tuple, list)) and len(common.SPIDER_SLEEP_TIME) == 2:
				sleep_times = random.randint(common.SPIDER_SLEEP_TIME[0], common.SPIDER_SLEEP_TIME[1])
				time.sleep(sleep_times)
			else:
				time.sleep(common.SPIDER_SLEEP_TIME)


class SpiderController(BaseController):

	def __init__(self, collector, request_cache):
		super(BaseController, self).__init__()
		self._thread_stop = False
		self._parser = []

		self._collector = collector
		self._request_cache = request_cache

	def run(self):
		while not self._thread_stop:
			requests = self._collector.get_requests(common.SPIDER_TASK_COUNT)
			if not requests:
				self.has_task_flag = False
			else:
				self.has_task_flag = True
				self.deal_requests(requests)

	def deal_requests(self, requests):
		for request in requests:
			response = None
			del_request_redis_after_request_to_db = False
			try:
				request_redis = request["request_redis"]
				request = request["request_obj"]
			except Exception as e:
				log.error(e)

			for parser in self._parser:
				try:
					if parser.name == request.parser_name:

						# TODO: download_middleware处理
						response = request.get_response()
						# 判断是否有回调函数
						if request.callback:
							# 检查回调函数是否可用
							callback_parser = (
								request.callback
								if callable(request.callback)
								else tools.get_method(parser, request.callback)
							)
							results = callback_parser(request, response)
						else:
							results = parser.parser(request, response)

						if results and not isinstance(results, Iterable):
							raise Exception(f"{parser.name}.{request.callback}必须可迭代的返回值")

						for result in results or []:
							if isinstance(result, Request):
								result.parser_name = result.parser_name or parser.name
								# 如果是同步的callback, 将解析request对象添加到requests中
								if result.request_sync:
									requests.append(result)
								else:
									result.parser_name = request.parser_name
									self._request_cache.add_request(result)
							elif isinstance(result, Item):
								# 存入item缓存中
								# self._item_cache.put(result)
								pass

						del_request_redis_after_request_to_db = True
				except Exception as e:
					log.exception(e)

			# 删除redis中已处理的任务
			if request_redis:
				if del_request_redis_after_request_to_db:
					self._request_cache.add_del_request(request_redis)

		# 休眠
		if common.SPIDER_SLEEP_TIME:
			if isinstance(common.SPIDER_SLEEP_TIME, (tuple, list)) and len(common.SPIDER_SLEEP_TIME) == 2:
				sleep_times = random.randint(common.SPIDER_SLEEP_TIME[0], common.SPIDER_SLEEP_TIME[1])
				time.sleep(sleep_times)
			else:
				time.sleep(common.SPIDER_SLEEP_TIME)



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/08/26

Desc: fastspider核心代码, redis
"""
import time

import redis
from fastspider.utils.logger import log

from fastspider.settings import common


class RedisDB(object):

	def __init__(self, url=None, ip=None, port=None, password=None, db=1, max_connections=30, **kwargs):
		self._url = url or common.REDISDB_URL
		self._ip = ip or common.REDISDB_IP
		self._port = port or common.REDISDB_PORT
		self._password = port or common.REDISDB_USER_PASS
		self._db = port or common.REDISDB_DB
		self._max_connections = max_connections
		self._kwargs = kwargs

		self._client = None
		self._redis = None

		self.connect()

	def __repr__(self):
		if self._url:
			return f"<RedisDB url:{self._url}>"

		return f"<RedisDB host: {self._ip} port: {self._ip} password: {self._password}>"

	@property
	def _client(self):
		try:
			if not self._redis.ping():
				raise ConnectionError("链接redis失败, 请检测redis配置")
		except Exception as e:
			self._reconnect()

		return self._redis

	@_client.setter
	def _client(self, val):
		self._redis = val

	def connect(self):
		"""
			链接redis
		"""
		if not self._url:
			if not self._ip or not self._port:
				raise Exception("请在配置 REDISDB_IP, REDISDB_PORT")
			if self._ip and self._port:
				self._client = redis.StrictRedis(host=self._ip, port=self._port, db=self._db, password=self._password,
				                                 decode_responses=True, max_connections=self._max_connections,
				                                 **self._kwargs)
		else:
			self._client = redis.StrictRedis.from_url(url=self._url, decode_responses=True)

		return self._client

	def _reconnect(self):
		"""
			重新链接redis
		"""
		retry_count = 0
		while True:
			try:
				retry_count += 1
				log.error(f"redis 连接断开, 重新连接 {retry_count}")
				if self.connect():
					log.info(f"redis 连接成功")
					return True
			except (ConnectionError, TimeoutError) as e:
				log.error(f"连接失败 e: {e}")

			time.sleep(2)

	def set(self, key, value, **kwargs):
		"""
			redis-string类型
		"""
		return self._client.set(key, value, **kwargs)

	def set_expire(self, key, seconds):
		"""
			设置过期时间
		"""
		return self._client.expire(key, seconds)

	def zadd(self, table_name, requests, priority=0):
		"""
			使用有序set集合进行数据的存储
		:param table_name: 集合名称
		:param requests:  请求支持list 或者单个值
		:param priority:  优先等级，支持list 或者 单个值, 可不传, 不传的情况下默认请求的优先级为0。或者也可以传入列表, 根据列表中的值来定义请求的执行优先级, 值越低, 越优先。
		:return:
		"""
		if isinstance(requests, list):
			if not isinstance(priority, list):
				priority = [priority] * len(requests)
			else:
				assert len(priority) == len(requests), "请求队列和优先级的值需要一一对应"

			# 批量操作
			pipeline = self._client.pipeline()
			pipeline.multi()
			for key, value in zip(requests, priority):
				pipeline.execute_command("ZADD", table_name, value, key)
				log.info(f"RedisDB 中插入任务成功, 数据格式为: {key}")
			return pipeline.execute()
		else:
			log.info(f"RedisDB 中插入任务成功, 数据格式为: {requests}")
			return self._client.execute_command("ZADD", table_name, priority, requests)

	def zrem(self, table_name, values):
		"""
			移除有序set集合中的元素, 如果元素不存在则忽略
		:param table_name: 集合名称
		:param values: 移除的元素, 支持列表 或者 单个值
		:return:
		"""
		if isinstance(values, list):
			self._client.zrem(table_name, *values)
		else:
			self._client.zrem(table_name, values)

	def zcount(self, table_name, priority_min=None, priority_max=None):
		"""
			计算有序set集合中的元素的个数
		:param table_name: 集合名称
		:param priority_min: 优先级范围的最小值
		:param priority_max: 优先级范围的最大值
		:return:
		"""
		if priority_min != None and priority_max != None:
			return self._client.zcount(table_name, priority_min, priority_max)
		else:
			return self._client.zcard(table_name)

	def zrangebyscore_set_score(
			self, table, priority_min, priority_max, score, count=None
	):
		"""
		@summary: 返回指定分数区间的数据 闭区间， 同时修改分数
		---------
		@param table: 集合名称
		@param priority_min: 最小分数
		@param priority_max: 最大分数
		@param score: 分数值
		@param count: 获取的数量，为空则表示分数区间内的全部数据
		---------
		@result:
		"""

		# 使用lua脚本， 保证操作的原子性
		lua = """
			-- local key = KEYS[1]
			local min_score = ARGV[1]
			local max_score = ARGV[2]
			local set_score = ARGV[3]
			local count = ARGV[4]
			
			-- 取值
			local datas = nil
			if count then
				datas = redis.call('zrangebyscore', KEYS[1], min_score, max_score, 'withscores','limit', 0, count)
			else
				datas = redis.call('zrangebyscore', KEYS[1], min_score, max_score, 'withscores')
			end
			
			local real_datas = {} -- 数据
			--修改优先级
			for i=1, #datas, 2 do
				local data = datas[i]
				local score = datas[i+1]
				
				table.insert(real_datas, data) -- 添加数据
				
				redis.call('zincrby', KEYS[1], set_score - score, datas[i])
			end
			
			return real_datas
		
		"""
		cmd = self._client.register_script(lua)
		if count:
			res = cmd(keys=[table], args=[priority_min, priority_max, score, count])
		else:
			res = cmd(keys=[table], args=[priority_min, priority_max, score])

		return res

	def zremrangebyscore(self, table_name, priority_min, priority_max):
		"""
			用于移除有序set集中，指定分数（score）区间内的所有成员
		:param table_name: 集合名称
		:param priority_min: 优先级最小值
		:param priority_max: 优先级最大值
		"""
		return self._client.zremrangebyscore(table_name, priority_min, priority_max)



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/09/02

Desc: fastspider核心代码, 定时器scheduler
"""
import threading

from fastspider.http.request.request import Request

from fastspider.cache.request_cache import RequestCache
from fastspider.core.controller.collector import Collector
from fastspider.core.controller.spider_controller import SpiderController
from fastspider.settings import common
from fastspider.utils.logger import log
from fastspider.utils import tools


class Scheduler(threading.Thread):
	__common_settings__ = {}

	def __init__(self, redis_key=None, thread_count=None, check_task_interval=None):
		"""
			定时执行爬虫任务的调度器
		"""
		self._thread_count = thread_count or common.SPIDER_THREAD_COUNT
		self._check_task_interval = check_task_interval
		self._parsers = []
		self._parsers_controller = []
		self._all_mission_done = False

		super(Scheduler, self).__init__()

		# 解析自定义的字典__common_settings__ 变量中的参数
		for k, v in self.__class__.__common_settings__.items():
			setattr(common, k, v)

		self._redis_key = redis_key or common.REDIS_KEY
		self._request_cache = RequestCache(self._redis_key)
		self._collector = Collector(redis_key)

	def __repr__(self):
		return f"<{self.name} 线程name: {threading.currentThread().name} 线程ID: {threading.currentThread().ident}>"

	def _start(self):
		# 启动相关线程
		# 1. 总请求任务/请求失败任务/请求删除任务 与 Redis的同步 线程
		self._request_cache.start()
		# 2. 获取Redis中任务请求request 并 写入到_todo_requests中的 线程
		self._collector.start()

		for i in range(self._thread_count):
			parser_controller = SpiderController(
				collector=self._collector,
				request_cache=self._request_cache
			)
			for parser in self._parsers:
				parser_controller.add_parser(parser)
			# 启动线程
			parser_controller.start()

			self._parsers_controller.append(parser_controller)

		# 3. 下发任务
		# TODO: 可能多进程之间会添加重复的任务, 下发任务处将使用redis 锁进行处理。暂时先考虑启动单进程处理任务, 不考虑锁的问题。
		self._add_task()

	def _add_task(self):
		"""
			添加待执行任务
		"""
		todo_requests_count = self._collector.get_requests_count()
		if todo_requests_count:
			log.info(f"检测有待执行任务 {todo_requests_count} 条, 不重新下发新任务, 将接着上次异常终止处继续执行抓取任务")
		else:
			for parser in self._parsers:
				for request in parser.start_requests():
					if request and not isinstance(request, Request):
						raise Exception("返回参数错误, 当前只支持yield fastspider.Request对象")
					request.parser_name = request.parser_name or parser.name
					self._request_cache.add_request(request)

		# 刷新一波待执行任务
		self._request_cache.flush()

	def all_thread_is_done(self):
		"""
			检测所有的线程是否都执行完毕, 如果执行完毕,则返回True。 如果其中任何一个线程没有执行完毕, 则返回Fasle
		"""
		for i in range(5):
			# 检查 _collector 状态
			if self._collector.is_collector_task() or self._collector.get_requests_count() > 0:
				return False
			# 检查每个线程中任务的执行状态,
			for parser in self._parsers_controller:
				if parser.has_task():
					return False

			# 检查 _request_cache 状态
			if self._request_cache.get_request_count() > 0 or self._request_cache.is_adding_request_to_db():
				return False

			tools.sleep_time(1)
		return True

	def all_thread_stop(self):
		"""
			手动暂停所有的线程
		"""
		self._request_cache.stop()

		self._collector.stop()

		for parser in self._parsers_controller:
			parser.stop()

	def start_spider(self):
		"""
			启动爬虫, 记录爬虫的开始时间等相关记录
		"""

		pass



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/07/29

Desc: fastspider核心代码, 合理调度redis中的requests队列中的任务
"""
import collections
import threading

from fastspider.http.request.request import Request

from fastspider.db.redis_db import RedisDB
from fastspider.settings import common
from fastspider.utils import tools
from fastspider.utils.logger import log


class Collector(threading.Thread):

	def __init__(self, redis_key=None):
		super(Collector, self).__init__()
		self._thread_stop = False

		self._redis_key = redis_key or common.REDIS_KEY
		self._interval = common.COLLECTOR_SLEEP_TIME
		self._get_task_count = common.COLLECTOR_TASK_COUNT
		# TODO: 后期修改为分布式后, _spider_mark可修改成不同主机的 ”IP-当前时间“
		self._spider_mark = f"spiders-{tools.get_current_timestamp()}"
		# 第一次获取任务
		self._first_get_task = True

		self._is_collector_task = False

		# 待执行请求队列
		self._todo_requests = collections.deque()

		self._redis_db = RedisDB()
		self._mission_request_name = common.REDIS_MISSION_REQUESTS.format(redis_key=redis_key)
		self._mission_spider_status = common.REDIS_SPIDER_STATUS.format(redis_key=redis_key)

	def _put_data(self):
		"""
			向待执行任务队列_todo_requests中插入任务
			# TODO
		:return:
		"""
		current_timestamp = tools.get_current_timestamp()
		if len(self._todo_requests) > self._get_task_count:
			return

		spider_count = self._redis_db.zcount(self._mission_spider_status,
		                                     priority_min=current_timestamp - (self._interval + 10),
		                                     priority_max=current_timestamp)

		request_count = self._get_task_count
		if spider_count:
			all_spider_count = self._redis_db.zcount(self._mission_spider_status)
			if all_spider_count % spider_count == 0:
				request_count = all_spider_count // spider_count
			else:
				request_count = all_spider_count // spider_count + 1

		request_count = (
			request_count
			if request_count <= self._get_task_count
			else self._get_task_count
		)

		# 第一次执行
		if self._first_get_task and spider_count <= 1:
			datas = self._redis_db.zrangebyscore_set_score(
				self._mission_request_name,
				priority_min=current_timestamp,
				priority_max=current_timestamp + common.REQUEST_LOST_TIMEOUT,
				score=300,
				count=None,
			)
			self._first_get_task = False
			lose_count = len(datas)
			if lose_count:
				log.info("重置丢失任务完毕，共{}条".format(len(datas)))

		# 取任务，只取当前时间戳以内的任务，同时将任务分数修改为 current_timestamp + setting.REQUEST_LOST_TIMEOUT
		requests_list = self._redis_db.zrangebyscore_set_score(
			self._mission_request_name,
			priority_min="-inf",
			priority_max=current_timestamp,
			score=current_timestamp + common.REQUEST_LOST_TIMEOUT,
			count=request_count,
		)

		if requests_list:
			self._is_collector_task = True
			# 存request
			self._put_requests(requests_list)

	def _put_requests(self, requests_list):
		for request in requests_list:
			try:
				request_dict = {
					"request_obj": Request.from_dict(eval(request)),
					"request_redis": request
				}
				self._todo_requests.append(request_dict)
			except Exception as e:
				log.error(f"数据格式转换失败, 失败原因: {e}, 失败请求request: {request}")

	def _add_node_heartbeat(self):
		"""
			添加节点的心跳
		"""
		self._redis_db.zadd(self._mission_spider_status, self._spider_mark, tools.get_current_timestamp())

	def _delete_node_heartbeat(self):
		"""
			删除节点的心跳
		"""
		self._redis_db.zremrangebyscore(self._mission_spider_status, "-inf",
		                                tools.get_current_timestamp() - (self._interval + 10))

	def run(self):
		while not self._thread_stop:
			try:
				# self._add_node_heartbeat()
				self._put_data()
			except Exception as e:
				log.error(f"{e}")

			self._is_collector_task = False
			tools.sleep_time(self._interval)

	def get_requests(self, count):
		"""
			获取指定个数的请求
		:param count: 请求个数
		:return: 获取到指定个数的请求
		"""
		requests = []
		count = count if count <= len(self._todo_requests) else len(self._todo_requests)
		while count:
			requests.append(self._todo_requests.popleft())
			count -= 1

		return requests

	def stop(self):
		self._thread_stop = True

	def get_requests_count(self):
		"""
			获取待处理任务数量
		"""
		return len(self._todo_requests) or self._redis_db.zcount(self._mission_request_name) or 0

	def is_collector_task(self):
		"""
			获取任务处理状态
		"""
		return self._is_collector_task



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/09/15

Desc: fastspider的预警功能, 对接飞书的机器人进行信息推送
"""
import base64
import json
import time
import hmac
from hashlib import sha256

import requests

from fastspider.settings.common import FEISHU_WEB_HOOK, FEISHU_SECRET


class FeiShuRobot(object):

	def __init__(self, webhook=None, secret=None):
		self._webhook = webhook or FEISHU_WEB_HOOK
		self._secret = secret or FEISHU_SECRET
		self._now_time = str(round(time.time()))
		self._key = f"{self._now_time}\n{self._secret}".encode("utf-8")

	def sign(self):
		code = hmac.new(self._key, "".encode("utf-8"), digestmod=sha256).digest()
		sign = base64.b64encode(code).decode("utf-8")
		return sign

	def send_msg(self, msg):
		sign = self.sign()

		data = {
			"timestamp": self._now_time,
			"sign": sign,
			"msg_type": "text",
			"content": {
				"text": msg
			}
		}
		headers = {
			"Content-Type": "application/json"
		}
		req = requests.post(self._webhook, headers=headers, data=json.dumps(data))

		return req



import logging
import os
import sys
from logging.handlers import RotatingFileHandler

import loguru

from fastspider.settings.common import LOGGER

LOG_FORMAT = "%(threadName)s|%(asctime)s|%(filename)s|%(funcName)s|line:%(lineno)d|%(levelname)s| %(message)s"


class LoggerHandler(logging.Handler):
	def emit(self, record):
		"""
			使用python3中的loguru日志模块
		"""
		# 一定会被执行的方法
		logger = loguru.logger.opt(depth=6, exception=record.exc_info)
		logger.log(record.levelname, record.getMessage())


def getlogger(name=None, path=None, log_level=None, write_to_file=None, write_to_console=None, file_number=None,
              max_bytes=None, encoding=None):
	"""
		获取定义的日志logger
	:param name: 日志名称, 默认为 fastspider
	:param path: 日志存储地址
	:param log_level: 日志级别 INFO DEBUG WARING ERROR
	:param write_to_console: 是否打印到控制台, True or False
	:param max_bytes: 日志文件存储的最大字节数, 默认是5M
	:param file_number: 日志文件最大个数, 默认是5个
	:param encoding: 编码格式 utf-8
	:return:
	"""
	name = name or LOGGER.get("LOG_NAME") or "fastspider"
	path = path or LOGGER.get("LOG_PATH") or "log/%s.log" % name
	log_level = log_level or LOGGER.get("LOG_LEVEL") or "DEBUG"
	write_to_file = write_to_file or LOGGER.get("LOG_IS_WRITE_TO_FILE") or False
	write_to_console = write_to_console or LOGGER.get("LOG_IS_WRITE_TO_CONSOLE") or True
	file_number = file_number or LOGGER.get("LOG_FILE_MAX_NUMBER") or "5"
	max_bytes = max_bytes or LOGGER.get("LOG_MAX_BYTES") or 5 * 1024 * 1024
	encoding = encoding or LOGGER.get("LOG_ENCODING") or "utf8"

	# 获取logger对象
	logger = logging.getLogger(name)
	logger.setLevel(log_level)
	# 获取格式化formatter
	formatter = logging.Formatter(LOG_FORMAT)

	if write_to_file:
		# 写入到文件
		if path and not os.path.exists(os.path.dirname(path)):
			os.mkdir(os.path.dirname(path))

		rotating_handler = RotatingFileHandler(
			path,
			mode='w',
			maxBytes=max_bytes,
			backupCount=file_number,
			encoding=encoding,
		)
		rotating_handler.setFormatter(formatter)
		logger.addHandler(rotating_handler)

	if write_to_console:
		# 打印到控制台
		loguru_handler = LoggerHandler()
		loguru_handler.setFormatter(formatter)
		logger.addHandler(loguru_handler)
	else:
		stream_handler = logging.StreamHandler()
		stream_handler.stream = sys.stdout
		stream_handler.setFormatter(formatter)
		logger.addHandler(stream_handler)

	return logger


class FLog(object):
	log = None

	def __getattr__(self, item):
		if self.__class__.log is None:
			self.__class__.log = getlogger()
		return getattr(self.__class__.log, item)


log = FLog()



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/07/27

Desc: 解析启动命令行参数, 并创建对应的爬虫
"""
import argparse

from fastspider.commands.create.create_spider import CreateFastSpider
from fastspider.commands.crawl.run_spider import RunFastSpider


def create():
	parser = argparse.ArgumentParser(description="启动命令参数解析")
	parser.add_argument(
		"-s",
		"--spider",
		nargs="+",
		help="创建爬虫\n"
		     "如 fastspider startspider -s <spider_name> <spider_type> "
		     "spider_type=1     表示创建轻量爬虫LightSpider; "
		     "spider_type=2     分布式海量数据抓取爬虫Spider; ",
		metavar=""
	)

	args = parser.parse_args()

	if args.spider:
		spider_name, *spider_type = args.spider
		if not len(spider_type):
			spider_type = "1"
		elif len(spider_type) == 1:
			spider_type = spider_type[0]
		else:
			raise Exception('spider_type error, must choice "1" "2" ')

		CreateFastSpider().create(spider_name, spider_type)


def crawl():
	# 运行爬虫
	parser = argparse.ArgumentParser(description="启动命令参数解析")
	parser.add_argument(
		"-s",
		"--spider",
		nargs="+",
		help="运行爬虫\n"
		     "如 fastspider crawl -s <spider_path> -c <thread_count>"
		     "spider_path           表示启动的爬虫的路径"
		     "thread_count          表示启动的爬虫线程个数",
		metavar=""
	)
	parser.add_argument(
		"-c",
		"--thread_count",
		nargs="+",
		help="爬虫的线程个数\n"
		     "如 fastspider crawl -s <spider_path> -c <thread_count>"
		     "spider_path           表示启动的爬虫的路径"
		     "thread_count          表示启动的爬虫线程个数",
		metavar=""
	)

	args = parser.parse_args()
	if args.spider:
		spider_path, *spider_params = args.spider
		if len(spider_params):
			raise Exception(
				"spider crawl error, command example fastspider crawl -s <spider_name> -c <thread_count> / fastspider crawl -s <spider_name>")

		thread_count, *spider_params = args.thread_count

		if isinstance(thread_count, str):
			try:
				thread_count = int(thread_count)
			except Exception as e:
				raise Exception("spider thread_count must int type")
			if thread_count > 20:
				raise Exception("spider thread_count must lt 20")

		if len(spider_params):
			raise Exception(
				"spider crawl error, command example fastspider crawl -s <spider_name> -c <thread_count> / fastspider crawl -s <spider_name>")

		# 启动爬虫
		RunFastSpider().run(spider_path, thread_count)



# encoding=utf-8
import json
import re
import time

import requests

import fastspider
from fastspider.utils.tools import cookies2dict


class TestSpider(fastspider.LightSpider):
	start_urls = [
		# "https://mdskip.taobao.com/mobile/queryH5Detail.htm?itemId=646958187838"
		"https://detail.tmall.com/item.htm?id=646958187838"
	]

	def start_requests(self):
		cookies = cookies2dict("_tb_token_=e6b5a0e9e3a75; cookie2=199d289165818892c52422ad65a04320; t=d4fea93bd0a64ba2f26f6c098b82f5b6")
		headers = {
			"referer": "https://detail.tmall.com/",
			"user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.107 Safari/537.36",
			"Accept": "*/*",
			"Accept-Encoding": "gzip, deflate, br",
			"Connection": "keep-alive"
		}
		for url in self.start_urls:
			meta = {
				"cookies": cookies,
				"headers": headers
			}
			yield fastspider.Request(url=url, callback=self.parse, headers=headers, meta=meta)

	def parse(self, request, response):
		meta = response.meta
		print(meta)
		text = response.content
		data = json.loads(str(text, encoding='gbk').replace('\n', ''))
		print(data)



if __name__ == "__main__":
	TestSpider().start()



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/08/05

Desc: pipeline基类

多态性是指具有不同功能的函数可以使用相同的函数名，这样就可以用一个函数名调用不同内容的函数.
"""
import abc
from typing import List, Dict


class BasePipeline(metaclass=abc.ABCMeta):
	"""
		使用ABCMeta作为元类来定义抽象基类
	"""

	@abc.abstractmethod
	def save_items(self, table, items: List[Dict]):
		"""
			保存数据
		:param table: 表名
		:param items: 数据 格式为 [{},{},{}....]
		:return: 保存成功返回True   保存失败返回False
		"""

		return True

	def update_items(self, table, items: List[Dict]):
		# TODO: 暂时不实现更新方法
		return True



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/09/02

Desc: fastspider核心爬虫CycleSpider的基类代码
"""
from fastspider.core.base.light_base import LightBase


class CycleBase(LightBase):

	def add_task(self):
		"""
			添加任务
		"""
		pass

	# def start_requests(self, task):
	# 	"""
	# 		启动任务, task为任务信息
	# 	"""
	# 	pass





from fastspider.core.spiders.light_spider import LightSpider
from fastspider.core.spiders.spider import Spider
from fastspider.http.request.request import Request

__all__ = [
	"LightSpider",
	"Request",
	"Spider",
]



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/08/05

Desc: mysql的pipeline管道
"""
from fastspider.pipeline.base_pipeline import BasePipeline


class MysqlPipeline(BasePipeline):

	def __init__(self):
		pass

	def save_items(self, table, items):
		"""
			管道pipeline中必须被实现的方法
		"""
		pass



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/07/29

Desc: fastspider核心代码, item实体缓存
"""
import importlib
from queue import Queue
from threading import Thread

from fastspider.item.item import Item
from fastspider.pipeline.base_pipeline import BasePipeline
from fastspider.utils import tools
from fastspider.settings import common


class ItemCache(Thread):
	MAX_ITEM_COUNT = 5000
	DEAL_ITEM_MAX_SIZE = 500

	def __init__(self, redis_key):
		super(ItemCache, self).__init__()

		self._thread_stop = False
		self._is_doing = False

		self._item_queue = Queue(maxsize=self.__class__.MAX_ITEM_COUNT)

		self._redis_key = redis_key
		# 定义Item名在redis中的key值
		self._table_item = "{redis_key}:{table_name}"
		# 存储不同Item对redis中的key值的对应关系
		self._item_table = {}

		self._pipelines = self.load_pipelines()

	def flush(self):
		"""
			从队列中获取指定数量的item对象, 如果item对象获取满500个, 则开始处理item对象. 如果item对象不足500个, 则单独处理item对象
		:return:
		"""
		count, items = 0, []

		while not self._item_queue.empty():
			data = self._item_queue.get_nowait()
			count += 1

			if isinstance(data, Item):
				items.append(data)
			# TODO: 暂时不做item去重
			else:
				# TODO: 其他情况, 暂不处理
				print("其他情况, 暂不处理")

			if count >= self.__class__.DEAL_ITEM_MAX_SIZE:
				self.__add_item_to_db(items)

				count, items = 0, []
		if items:
			self.__add_item_to_db(items)

	def __pick_items(self, items):
		"""
			将不同Item对象区分开
		"""
		item_dict = {}
		while items:
			item = items.pop(0)
			item_name = item.item_name
			# 将table_name修改为小写命名规范，如DoubanItem 修改为douban_item
			item_name = tools.pascal_case_to_snake_case(item_name)

			if self._item_table.get(item_name):
				table_item = self._item_table.get(item_name)
			else:
				# 组装redis中存储的key值
				table_item = self._table_item.format(
					redis_key=self._redis_key, table_name=item_name
				)
				self._item_table[item_name] = table_item

			if table_item not in item_dict:
				item_dict[table_item] = []
			# 将不同Item对象的数据保存到item_dict字典中.格式为{item对应的redis key值: [item字典, item字典].....}
			item_dict[table_item].append(item.to_dict)
		return item_dict

	def load_pipelines(self):
		"""
			加载settings中定义的piplines
		:return:
		"""
		pipelines = []
		for pipeline in common.ITEM_PIPELINES:
			module, class_name = pipeline.rsplit(".", 1)
			pipeline_cls = importlib.import_module(module).__getattribute__(class_name)

			if not isinstance(pipeline_cls(), BasePipeline):
				raise Exception(f"{class_name}必须继承 BasePipeline ")

			pipelines.append(pipeline_cls())
		return pipelines

	def __save_item_to_db(self, table_name, item_data, is_update=False):
		"""
			保存数据
		:param table_name: redis key名称
		:param item_data: 存储的数据
		:param is_update: True为更新数据  False为插入数据
		:return:
		"""

		for pipeline in self._pipelines:
			if is_update:
				# TODO: 更新数据
				pass
			else:
				if not pipeline.save_items(table_name, item_data):
					print("插入数据失败")

	def __add_item_to_db(self, items):
		"""
			处理item对象
		"""
		self._is_doing = True
		# 处理item对象的信息, 返回redis队列和序列化item信息
		items_dict = self.__pick_items(items)
		if items_dict:
			table_name, item_data = items_dict.popitem()
			self.__save_item_to_db(table_name, item_data)

		# 是否正在执行存储操作
		self._is_doing = False

	def put(self, item):
		"""
			将item对象存在队列中
		:param item: Item对象
		"""
		self._item_queue.put(item)

	def run(self):
		while not self._thread_stop:
			self.flush()
			tools.sleep_time(0.5)

	def stop(self):
		self._thread_stop = True

	def is_empty(self):
		"""
			判断item队列是否为空
		:return: 返回True 为空, 返回False 不为空
		"""
		return self._item_queue.empty()

	def adding_item_to_db(self):
		"""
			判断item是否正在存储数据库中
		:return:
		"""
		return self._is_doing



# encoding=utf-8
"""
Auth: coco369
Email: 779598160@qq.com

CreateTime: 2021/08/03

Desc: fastspider核心代码, 封装webdriver
"""
import threading
from queue import Queue

from selenium import webdriver
from selenium.webdriver.chrome.webdriver import WebDriver

from fastspider.http import user_agent
from fastspider.settings import common

from fastspider.utils.tools import Singleton


class FastWebDriver(WebDriver):
	# TODO: 2021/08/03 目前暂时只支持Chrome
	CHROME = "Chrome"

	def __init__(self, driver_type=CHROME, not_load_images=True, user_agent=None, proxies=None, timeout=30,
	             windows_size=(1024, 800), executable_path=None, headless=None, **kwargs):

		self._driver_type = driver_type
		self.not_load_images = not_load_images
		self._user_agent = user_agent
		self._proxies = proxies
		self._timeout = timeout
		self._window_size = windows_size
		self._executable_path = executable_path
		self._headless = headless

		if self._driver_type == FastWebDriver.CHROME:
			self.driver = self.chrome_driver()

		# 当网络差get(url)不返回,但也不报错, 设置超时选项来解决程序卡住的问题。
		self.driver.set_page_load_timeout(self._timeout)
		# 设置10秒脚本超时时间
		self.driver.set_script_timeout(self._timeout)

	def chrome_driver(self):
		"""
			获取chrome浏览器对象
		:return: 浏览器对象
		"""
		options = webdriver.ChromeOptions()
		options.add_experimental_option("excludeSwitches", ["enable-automation"])
		options.add_experimental_option("useAutomationExtension", False)

		options.add_argument("--no-sandbox")
		options.add_argument("--disable-dev-shm-usage")

		# 加载图片设置
		if self.not_load_images:
			prefs = {"profile.managed_default_content_settings.images": 2}
			options.add_experimental_option("prefs", prefs)

		# user-agent设置
		if not self._user_agent:
			random_user_agent = user_agent.get_ua(common.USER_AGENT_TYPE)
			options.add_argument(f"user-agent={random_user_agent}")
		else:
			options.add_argument(f"user-agent={self._user_agent}")

		# 代理proxies设置
		if self._proxies:
			options.add_argument(f"--proxy-server={self._proxies}")

		# 无头模式设置
		if self._headless:
			options.add_argument("--headless")

		# 浏览器窗口大小设置
		if self._window_size:
			options.add_argument(f"--window-size={self._window_size[0]},{self._window_size[1]}")

		# 是否指定webdriver.exe的路径设置
		if self._executable_path:
			driver = webdriver.Chrome(options=options, executable_path=self._executable_path)
		else:
			driver = webdriver.Chrome(options=options)
		# execute_cdp_cmd 通过调用DevTools来调试
		# 模拟浏览器访问被识别, 应对方案如下
		driver.execute_cdp_cmd(
			"Page.addScriptToEvaluateOnNewDocument",
			{
				"source": """
	                Object.defineProperty(navigator, 'webdriver', {
	                    get: () => undefined
	                })
                """
			},
		)
		return driver

	@property
	def cookies(self):
		new_cookies = {}
		for cookie in self.get_cookies():
			new_cookies[cookie["name"]] = cookie["value"]
		return new_cookies

	def __getattr__(self, name):
		"""
			WebDriver启动后, 会生成相应的控制浏览器会话的session_id等参数。
			其父类WebDriver需要使用到session_id等属性, 因此通过当前浏览器对象this.driver将相关session_id等属性获取到, 并作为属性的value值返回
		:param name: 参数
		:return: 返回当前浏览器对象driver的相关属性
		"""
		if self.driver:
			return getattr(self.driver, name)
		else:
			raise AttributeError


@Singleton
class WebDriverPool:
	"""
		webdriver池
	"""

	def __init__(self, pool_count=1, **kwargs):
		self.pool_count = pool_count
		self.queue = Queue(maxsize=self.pool_count)
		self.kwargs = kwargs
		# 多重锁, RLock允许在同一线程中被多次acquire, acquire和release必须成对出现
		self.thread_lock = threading.RLock()
		self.queue_count = 1

	def get(self):
		"""
		:return: 获取浏览器deiver对象
		"""
		# TODO: queue中为什么需要5个driver对象
		while self.queue_count <= self.queue.maxsize:
			with self.thread_lock:
				driver = FastWebDriver(**self.kwargs)
				self.queue.put(driver)
				self.queue_count += 1

		driver = self.queue.get()
		return driver

	def put(self, driver):
		self.queue.put(driver)

	def remove(self, driver):
		driver.close()
		self.queue_count -= 1

	def close(self):
		while not self.queue.empty():
			driver = self.queue.get()
			driver.close()
			self.queue_count -= 1
